/**************************************************************************/
/* HandleRexx.c - ARexx message processing                                */
/**************************************************************************/

#include "FetchRefs.h"

char *RexxHostName;
LONG GotoLine;

void
ARexxLoop(void)
{
    /* ARexx loop */
    for(;;)
    {
        LONG mask;

        mask = Wait(SIGBREAKF_CTRL_C | (1 << RexxSigBit));

        /* FR_QUIT will signal FindTask(NULL) with SIGBREAKF_CTRL_C */
        if (mask & SIGBREAKF_CTRL_C)
            return;

        if (mask & (1 << RexxSigBit))
            ProcessRexxCommands(NULL);
    }
}

/* The theory of parsing the messages is based on a two pass ReadArgs() parsing.
 *
 * First the entire message is parsed against RexxArgStr, and as all commands
 * are postfixed by /F the entire rest of the line is considered to be the
 * parameter of this single command. If an unrecognised keyword is entered then
 * everything is considered a parameter to DUMMY (as it's the first one); thus
 * out-of-context commands are flushed out here.
 *
 * Having had ReadArgs() put everything but the command name into a string, we
 * can parse this against a template which differs for each command, but as all
 * the templates are in an array we simply need one ReadArgs() call along with
 * the right offset into this array (which is determined by what the command
 * was). After having parsed again (but only if it's a succes), a switch can
 * decide what actions should be made with the parameters parsed last - which
 * depends on what the first parsing resulted in.
 *
 * As bright as oil, right? Anyway, I like it :-).
 *
 * (And later I learned about ARexxBox... :-).
 */

/* Templates used by ARexx interface */
const UBYTE RexxArgStr[] =
    "DUMMY/F,FR_QUIT/F,FR_GET/F,FR_CLEAR/F,FR_NEW/F,FR_ADD/F";

const UBYTE (*RexxTemplates[])[] = {
/* DUMMY    */  "",
/* FR_QUIT  */  "",
/* FR_GET   */  "FIND/A,TO/A,PUBSCREEN,FILEREF/S,CASE/S",
/* FR_CLEAR */  "",
/* FR_NEW   */  "FILE/M",
/* FR_ADD   */  "FILE/M"
};

/* The following routine is called by DICE everytime I receive an ARexx msg.
 * Depending on what the message was, different actions are performed.
 */
LONG
DoRexxCommand(void *msg, struct MsgPort *port, char *arg0, char **pres)
{
    LONG rc = 0;
    struct RDArgs *RexxArgs, *ArgsArgs;
    LONG RexxResult[sizeof(RexxTemplates) / sizeof(RexxTemplates[0])];
    LONG ArgsResult[10];
    UBYTE str[256];

    /* Flush out garbage as ReadArgs() allows for pre-initialized values */
    clrmem(RexxResult, sizeof(RexxResult));
    clrmem(ArgsResult, sizeof(ArgsResult));

    /* ReadArgs() needs a '\n' at the end of its strings :-( */
    strcpy(str, arg0);
    strcat(str, "\n");

    if ((RexxArgs = AllocDosObject(DOS_RDARGS, NULL)) && (ArgsArgs = AllocDosObject(DOS_RDARGS, NULL)))
    {
        RexxArgs->RDA_Source.CS_Buffer = str;
        RexxArgs->RDA_Source.CS_Length = strlen(str);
        RexxArgs->RDA_Flags = RDAF_NOPROMPT;
        if (ReadArgs(RexxArgStr, RexxResult, RexxArgs))
        {
            LONG argno = 0;

            /* Figure out which command we got */
            while (!RexxResult[argno])
                argno++;

            /* If a template is specified: parse the arguments */
            if (strlen(RexxTemplates[argno]))
            {
                strcpy(str, (STRPTR)RexxResult[argno]);
                strcat(str, "\n");
                ArgsArgs->RDA_Source.CS_Buffer = str;
                ArgsArgs->RDA_Source.CS_Length = strlen(str);
                ArgsArgs->RDA_Flags = RDAF_NOPROMPT;
                if (!(ReadArgs(RexxTemplates[argno], ArgsResult, ArgsArgs)))
                    rc = IoErr();
            }

            if (rc == 0)
                switch (argno)
                {
                    case 0: rc = ERROR_LINE_TOO_LONG;
                            break;

                    case 1: Signal(FindTask(NULL), SIGBREAKF_CTRL_C);
                            break;

                    case 2:
                        SOpts.Reference = (STRPTR)ArgsResult[0];
                        SOpts.DestFile = (STRPTR)ArgsResult[1];
                        SOpts.PubScreen = (STRPTR)ArgsResult[2];
                        SOpts.FileRef = ArgsResult[3];
                        SOpts.Case = ArgsResult[4];
                        rc = FindRef();
                        break;

                    case 3:
                        FreeRefs();
                        break;

                    case 4:
                        FreeRefs(); /* FR_NEW falls through to FR_ADD */
                    case 5: 
                        { 
                            LONG count; STRPTR path;
                            
                            for (count = 0; path = ((UBYTE **)ArgsResult[0])[count]; count++)
                                ReadWild(path);
                        }
                        break;
                }
        } else
            rc = IoErr();

        FreeArgs(ArgsArgs);
        FreeDosObject(DOS_RDARGS, ArgsArgs);
    } else
        rc = ERROR_NO_FREE_STORE;

    if (RexxArgs)
    {
        FreeArgs(RexxArgs);
        FreeDosObject(DOS_RDARGS, RexxArgs);
    }

    switch (rc)
    {
        case 0:  SPrintf(str, "%ld", GotoLine);
                 break;

        case 1:  strcpy(str, "FetchRefs: Reference not found");
                 break;

        case 2:  strcpy(str, "FetchRefs: Aborted!");
                 break;

        default: Fault(rc, "FetchRefs", str, 256);
                 break;
    }

    SetRexxVar(msg, "RC2", str, strlen(str));
    return(rc ? RETURN_ERROR : RETURN_OK);
}

/* Low-cost SPrintf routine. Returns the length of the created string.
 * Seems like Commodore forgot to put this in ROM? sprintf from amiga.lib is
 * kind of hard to get to when c.lib is linked in first.
 */
static void SPrintfSupp(__D0 UBYTE Char, __A3 STRPTR *ptr);

LONG
SPrintf(STRPTR buf, STRPTR fmt, ...)
{
    va_list args;
    STRPTR start = buf;

    va_start(args, fmt);
    RawDoFmt(fmt, args, SPrintfSupp, &buf);
    va_end(args);
    return((LONG)buf - (LONG)start - 1);
}

/* Call back function for SPrintf. Is called to push each char. */
static void
SPrintfSupp(__D0 UBYTE Char, __A3 STRPTR *ptr)
{
    *(*ptr)++ = Char;
}

